@@ -47,4 +47,15 @@ static UV isqrt(UV n) {
   return root;
 }
 
+#ifdef FUNC_gcd_ui
+static UV gcd_ui(UV x, UV y) {
+  UV t;
+  if (y < x) { t = x; x = y; y = t; }
+  while (y > 0) {
+    t = y;  y = x % y;  x = t;  /* y1 <- x0 % y0 ; x1 <- y0 */
+  }
+  return x;
+}
+#endif
+
 #endif
